from typing import TypeVar, Callable, Tuple
from queue import Queue

from amino import do, Do, IO, List, Nil, Maybe, Nothing, Just, Map
from amino.logging import module_log
from amino.case import Case
from amino.lenses.lens import lens

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.nvim.io.state import NS
from ribosome.util.menu.prompt.data import (InputState, InputResources, InputChar, PrintableChar, SpecialChar,
                                            PromptInputAction, PromptInput, PromptInterrupt, Prompt, PromptConsumer,
                                            PromptEcho, PromptAction, PromptStateTrans, PromptQuit, PromptUnit,
                                            PromptUpdate, PromptUpdateChar, PromptConsumerInput, PromptUpdateConsumer,
                                            PromptInit, PromptUpdateInit, PromptConsumerUpdate, PromptConsumerInit,
                                            PromptConsumerChanged, PromptConsumerUnhandled, PromptConsumerLocal,
                                            PromptPassthrough)
from ribosome.nvim.api.command import nvim_atomic_commands
from ribosome.util.menu.prompt.input import input_loop
from ribosome.util.menu.prompt.interrupt import intercept_interrupt, stop_prompt, stop_prompt_s

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


@do(NS[InputState[A, B], C])
def prompt_state_fork(
        echo: Callable[[], NS[InputState[A, B], C]],
        passthrough: Callable[[], NS[InputState[A, B], C]],
) -> Do:
    state = yield NS.inspect(lambda a: a.state)
    yield (
        echo()
        if isinstance(state, PromptEcho) else
        passthrough()
    )


def prompt_state_fork_strict(echo: C, passthrough: C) -> NS[InputState[A, B], C]:
    return prompt_state_fork(
        lambda: NS.pure(echo),
        lambda: NS.pure(passthrough),
    )


def prompt_if_echoing(thunk: Callable[[], NS[InputState[A, B], C]]) -> NS[InputState[A, B], C]:
    return prompt_state_fork(thunk, lambda: NS.unit)


class process_char(Case[InputChar, NS[InputState[A, B], None]], alg=InputChar):

    def printable(self, a: PrintableChar) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(PromptUpdateChar(a)))

    def special(self, a: SpecialChar) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(PromptUpdateChar(a)))


def prompt_fragment(text: str, highlight: Maybe[str]) -> List[str]:
    hl = highlight.get_or_strict('None')
    return List(f'echohl {hl}', f'echon "{text}"')


@do(NS[InputState[A, B], None])
def redraw_prompt() -> Do:
    prompt, cursor = yield NS.inspect(lambda s: (s.prompt, s.cursor))
    pre, cursor_char, post = prompt.pre, prompt.post[:1], prompt.post[1:]
    cmds = List((pre, Nothing), (cursor_char, Just('RibosomePromptCaret')), (post, Nothing)).flat_map2(prompt_fragment)
    yield NS.lift(nvim_atomic_commands(cmds.cons('redraw')))


@do(NS[InputState[A, B], None])
def redraw_if_echoing() -> Do:
    state = yield NS.inspect(lambda a: a.state)
    yield redraw_prompt() if isinstance(state, PromptEcho) else NS.unit


def insert_char(char: str) -> Callable[[Prompt], Prompt]:
    def insert_char(prompt: Prompt) -> Prompt:
        log.debug(f'inserting {char} into prompt')
        return prompt.copy(cursor=prompt.cursor + 1, pre=prompt.pre + char)
    return insert_char


def backspace(prompt: Prompt) -> Prompt:
    log.debug('applying backspace to prompt')
    parts = Just((prompt.pre[-1:], prompt.pre[:-1])) if len(prompt.pre) > 0 else Nothing
    return parts.map2(lambda l, i: prompt.copy(pre=i, cursor=prompt.cursor - 1)).get_or_strict(prompt)


special_prompt_keys = Map({
    '<space>': insert_char(' '),
    '<bs>': backspace,
})


class update_prompt_text(Case[InputChar, Maybe[Prompt]], alg=InputChar):

    def __init__(self, prompt: Prompt) -> None:
        self.prompt = prompt

    def printable(self, a: PrintableChar) -> Prompt:
        return Just(insert_char(a.char)(self.prompt))

    def special(self, a: SpecialChar) -> Prompt:
        return special_prompt_keys.lift(a.char).map(lambda f: f(self.prompt))


@do(IO[List[PromptInputAction[A]]])
def dequeue(inputs: Queue) -> Do:
    first = yield IO.delay(inputs.get)
    tail = yield dequeue(inputs) if inputs.qsize() > 0 else IO.pure(Nil)
    return tail.cons(first)


class process_input(Case[PromptInputAction[A], Maybe[PromptUpdate[A]]], alg=PromptInputAction):

    def init(self, a: PromptInit[A]) -> Maybe[PromptUpdate[A]]:
        return Just(PromptUpdateInit())

    def input(self, a: PromptInput[A]) -> Maybe[PromptUpdate[A]]:
        return Just(PromptUpdateChar(a.char))

    def interrupt(self, a: PromptInterrupt[A]) -> Maybe[PromptUpdate[A]]:
        return Nothing

    def consumer(self, a: PromptConsumerInput[A]) -> Maybe[PromptUpdate[A]]:
        return Just(PromptUpdateConsumer(a.data))


class execute_prompt_action(Case[PromptAction, NS[InputResources[A, B], None]], alg=PromptAction):

    def state_trans(self, a: PromptStateTrans) -> NS[InputResources[A, B], None]:
        return NS.modify(lens.state.state.set(a.state))

    def quit(self, a: PromptQuit) -> NS[InputResources[A, B], None]:
        return stop_prompt_s()

    def unit(self, a: PromptUnit) -> NS[InputResources[A, B], None]:
        return NS.unit


class process_prompt_update_char(
        Case[PromptUpdate[B], NS[InputState[A, B], PromptConsumerUpdate[B]]],
        alg=PromptUpdate,
):

    def init(self, a: PromptUpdateInit[B]) -> NS[InputState[A, B], PromptConsumerUpdate[B]]:
        return NS.pure(PromptConsumerInit())

    def char(self, update: PromptUpdateChar[B]) -> NS[InputState[A, B], PromptConsumerUpdate[B]]:
        def apply_char(s: InputState[A, B]) -> NvimIO[Tuple[InputState[A, B], PromptConsumerUpdate[B]]]:
            return N.pure(update_prompt_text(s.prompt)(update.char).cata(
                lambda prompt: (s.set.prompt(prompt), PromptConsumerChanged()),
                lambda: (s, PromptConsumerUnhandled(update.char)),
            ))
        return prompt_state_fork(
            lambda: NS.apply(apply_char),
            lambda: NS.pure(PromptConsumerUnhandled(update.char)),
        )

    def consumer(self, a: PromptUpdateConsumer[B]) -> NS[InputState[A, B], PromptConsumerUpdate[B]]:
        return NS.pure(PromptConsumerLocal(a.data))


@do(NS[InputResources[A, B], None])
def process_prompt_update(process: PromptConsumer, action: PromptUpdate[B]) -> Do:
    consumer = yield process_prompt_update_char.match(action).zoom(lens.state)
    yield prompt_if_echoing(redraw_prompt).zoom(lens.state)
    prompt_action = yield process(consumer).zoom(lens.state)
    yield execute_prompt_action.match(prompt_action)


def update_prompt(process: PromptConsumer) -> Callable[[List[PromptUpdate[B]]], NS[InputResources[A, B], None]]:
    def update_prompt(chars: List[PromptUpdate[B]]) -> NS[InputState[A, B], None]:
        return chars.traverse(lambda char: process_prompt_update(process, char), NS).replace(None)
    return update_prompt


class process_action(Case[PromptUpdate[B], NS[InputState[A, B], None]], alg=PromptUpdate):

    def init(self, a: PromptUpdateInit[B]) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(a))

    def input_char(self, a: PromptUpdateChar[B]) -> NS[InputState[A, B], None]:
        return process_char.match(a.char)

    def custom(self, a: PromptUpdateConsumer[B]) -> NS[InputState[A, B], None]:
        return NS.modify(lambda s: s.append1.actions(a))


def pop_actions(s: InputState[A, B]) -> NvimIO[Tuple[InputState[A, B], List[PromptUpdate[B]]]]:
    return N.pure((s.set.actions(Nil), s.actions))


@do(NS[InputResources[A, B], None])
def prompt_recurse(input: List[PromptUpdate[B]]) -> Do:
    stop = yield NS.inspect(lambda a: a.stop)
    yield input.traverse(process_action.match, NS).zoom(lens.state)
    current = yield NS.apply(pop_actions).zoom(lens.state)
    update = yield NS.inspect(lambda s: s.update)
    yield update(current)
    yield NS.unit if stop.is_set() else prompt_loop()


@do(NS[InputResources[A, B], None])
def prompt_loop() -> Do:
    inputs = yield NS.inspect(lambda s: s.inputs)
    input = yield NS.from_io(dequeue(inputs))
    actions = input.traverse(process_input.match, Maybe)
    yield actions.cata_strict(prompt_recurse, NS.unit)


@do(NS[InputResources[A, B], None])
def start_prompt_loop() -> Do:
    inputs = yield NS.inspect(lambda s: s.inputs)
    yield NS.from_io(IO.delay(inputs.put, PromptInit()))
    yield prompt_loop()


@do(NvimIO[A])
def prompt(process: PromptConsumer, initial: A, start_insert: bool) -> Do:
    log.debug(f'running prompt with {initial}')
    state = PromptEcho() if start_insert else PromptPassthrough()
    res = InputResources.cons(
        InputState.cons(initial, state=state),
        update_prompt(process),
    )
    input_thread = yield N.fork(input_loop, res)
    result = yield intercept_interrupt(res.inputs, res.stop, res, start_prompt_loop().run_s(res))
    yield N.from_io(stop_prompt(res.inputs, res.stop))
    yield N.simple(input_thread.join)
    return result.state.data


__all__ = ('prompt', 'prompt_state_fork', 'prompt_state_fork_strict', 'prompt_if_echoing',)



from typing import TypeVar

from amino import do, Do, Map, List, Maybe
from amino.lenses.lens import lens
from amino.logging import module_log

from ribosome.util.menu.auto.data import AutoS
from ribosome.nvim.io.state import NS
from ribosome.util.menu.data import MenuPrompt, MenuUpdateCursor, MenuLine, visible_menu_indexes, MenuState, MenuQuit
from ribosome.util.menu.prompt.data import PromptPassthrough, InputState
from ribosome.util.menu.prompt.run import prompt_state_fork_strict

log = module_log()
A = TypeVar('A')
S = TypeVar('S')
ML = TypeVar('ML')
U = TypeVar('U')


def menu_cmd_esc() -> AutoS[S, ML, U]:
    return prompt_state_fork_strict(
        MenuPrompt(PromptPassthrough()),
        MenuQuit(),
    )


@do(NS[InputState[MenuState[S, ML, U], U], int])
def menu_direction() -> Do:
    bottom = yield NS.inspect(lambda a: a.data.config.bottom)
    return -1 if bottom else 1


@do(AutoS[S, ML, U])
def menu_cmd_scroll(offset: int) -> Do:
    direction = yield menu_direction()
    visible_indexes = yield NS.inspect(lambda a: visible_menu_indexes(a.data.content))
    count = visible_indexes.length
    yield NS.modify(lens.data.cursor.modify(lambda a: (a + direction * offset) % count))
    return MenuUpdateCursor()


def menu_cmd_up() -> AutoS[S, ML, U]:
    return menu_cmd_scroll(-1)


def menu_cmd_down() -> AutoS[S, ML, U]:
    return menu_cmd_scroll(1)


def toggle_selected(lines: List[MenuLine[A]], index: int) -> Maybe[List[MenuLine[A]]]:
    return lines.modify_at(index, lambda b: b.mod.selected(lambda c: not c))


visibility_error = 'broken visibility map for menu'


@do(AutoS[S, ML, U])
def menu_cmd_select_cursor() -> Do:
    cursor = yield NS.inspect(lambda a: a.cursor)
    content = yield NS.inspect(lambda a: a.data.content)
    visible_indexes = visible_menu_indexes(content)
    index = yield NS.from_maybe(visible_indexes.lift(cursor), lambda: visibility_error)
    toggled = yield NS.from_maybe(toggle_selected(content.lines, index), lambda: visibility_error)
    yield NS.modify(lens.data.content.lines.set(toggled))
    return MenuUpdateCursor()


@do(AutoS[S, ML, U])
def menu_cmd_select_all() -> Do:
    yield NS.modify(lens.data.content.lines.modify(lambda a: a.map(lambda b: b.mod.selected(lambda c: not c))))
    return MenuUpdateCursor()


builtin_mappings = Map({
    '<esc>': menu_cmd_esc,
    'q': menu_cmd_esc,
    'k': menu_cmd_up,
    '<c-k>': menu_cmd_up,
    'j': menu_cmd_down,
    '<c-j>': menu_cmd_down,
    '<space>': menu_cmd_select_cursor,
    '*': menu_cmd_select_all,
})

__all__ = ('builtin_mappings',)



from asyncio import BaseEventLoop, WriteTransport, SubprocessTransport, Transport
from threading import Thread
from concurrent.futures import Future

from amino import Dat, List, ADT, Maybe, Nothing, Just, Path
from amino.logging import module_log

from ribosome.rpc.concurrency import OnMessage, OnError
from ribosome.rpc.error import processing_error, RpcReadErrorUnknown

log = module_log()


class AsyncioPipes(ADT['AsyncioPipes']):
    pass


class AsyncioEmbed(AsyncioPipes):

    def __init__(self, proc: List[str]) -> None:
        self.proc = proc


class AsyncioStdio(AsyncioPipes):
    pass


class AsyncioSocket(AsyncioPipes):

    def __init__(self, path: Path) -> None:
        self.path = path


class AsyncioLoopThread(Dat['AsyncioLoopThread']):

    def __init__(self, thread: Maybe[Thread]) -> None:
        self.thread = thread

    def update(self, thread: Thread) -> None:
        self.thread = Just(thread)

    def reset(self) -> None:
        self.thread = Nothing


class AsyncioResources(Dat['AsyncioResources']):

    @staticmethod
    def cons(
            transport: Future,
    ) -> 'AsyncioResources':
        return AsyncioResources(
            transport,
        )

    def __init__(self, transport: Future) -> None:
        self.transport = transport


class Asyncio(Dat['Asyncio']):

    @staticmethod
    def cons(
            loop: BaseEventLoop,
            pipes: AsyncioPipes,
            resources: AsyncioResources,
            thread: AsyncioLoopThread=None,
    ) -> 'Asyncio':
        return Asyncio(loop, pipes, resources, thread or AsyncioLoopThread(Nothing))

    def __init__(
            self,
            loop: BaseEventLoop,
            pipes: AsyncioPipes,
            resources: AsyncioResources,
            thread: AsyncioLoopThread,
    ) -> None:
        self.loop = loop
        self.pipes = pipes
        self.resources = resources
        self.thread = thread


class EmbedProto(Dat['EmbedProto']):

    def __init__(self, asio: Asyncio, on_message: OnMessage, on_error: OnError) -> None:
        self.asio = asio
        self.on_message = on_message
        self.on_error = on_error

    def connection_made(self, transport: SubprocessTransport) -> None:
        self.asio.resources.transport.set_result(transport.get_pipe_transport(0))

    def connection_lost(self, exc: Exception) -> None:
        pass

    def pipe_connection_lost(self, fd: int, exc: Exception) -> None:
        pass

    def pipe_data_received(self, fd: int, data: bytes) -> None:
        if fd == 1:
            self.on_message(data).attempt.leffect(processing_error(data))
        else:
            self.on_error(RpcReadErrorUnknown(data or b'no error message')).attempt.leffect(processing_error(None))

    def process_exited(self) -> None:
        pass


class BasicProto(Dat['BasicProto']):

    def __init__(self, asio: Asyncio, on_message: OnMessage, on_error: OnError) -> None:
        self.asio = asio
        self.on_message = on_message
        self.on_error = on_error

    def connection_made(self, transport: Transport) -> None:
        try:
            if isinstance(transport, WriteTransport):
                self.asio.resources.transport.set_result(transport)
        except Exception as e:
            log.caught_exception(f'setting transport {transport}', e)

    def connection_lost(self, exc: Exception) -> None:
        pass

    def data_received(self, data: bytes) -> None:
        self.on_message(data).attempt.leffect(processing_error(data))


__all__ = ('EmbedProto', 'BasicProto', 'AsyncioPipes', 'AsyncioEmbed', 'AsyncioStdio', 'AsyncioSocket',
           'AsyncioLoopThread', 'AsyncioResources', 'Asyncio',)



from kallikrein import Expectation
from kallikrein.matchers.comparison import eq

from ribosome.test.klk.expectation import await_k_with
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.ui import current_cursor


def current_cursor_is(line: int, col: int) -> NvimIO[Expectation]:
    return await_k_with(eq((line, col)), current_cursor)


__all__ = ('current_cursor_is',)



from typing import Callable, Optional

from amino import ADT
from amino.logging import module_log

log = module_log()


class RpcReadError(ADT['RpcReadError']):
    pass


class RpcProcessExit(RpcReadError):
    pass


class RpcReadErrorUnknown(RpcReadError):

    def __init__(self, message: str) -> None:
        self.message = message


def processing_error(data: Optional[bytes]) -> Callable[[Exception], None]:
    def processing_error(error: Exception) -> None:
        if data:
            log.debug(f'{error}: {data}')
        log.error(f'error processing message from nvim: {error}')
    return processing_error


__all__ = ('RpcReadError', 'RpcProcessExit', 'RpcReadErrorUnknown', 'processing_error',)



import sys
from asyncio import get_child_watcher, run_coroutine_threadsafe
from typing import Callable, Coroutine

from amino import IO, do, Do, Try
from amino.case import Case
from amino.logging import module_log

from ribosome.rpc.concurrency import OnMessage, OnError
from ribosome.rpc.io.data import (AsyncioResources, AsyncioPipes, Asyncio, AsyncioEmbed, EmbedProto, AsyncioStdio,
                                  BasicProto, AsyncioSocket)

log = module_log()


class connect_asyncio(Case[AsyncioPipes, IO[None]], alg=AsyncioPipes):

    def __init__(self, asio: Asyncio, on_message: OnMessage, on_error: OnError) -> None:
        self.asio = asio
        self.on_message = on_message
        self.on_error = on_error

    @do(IO[Coroutine])
    def embed(self, a: AsyncioEmbed) -> Do:
        child_watcher = yield IO.delay(get_child_watcher)
        yield IO.delay(child_watcher.attach_loop, self.asio.loop)
        return self.asio.loop.subprocess_exec(lambda: EmbedProto(self.asio, self.on_message, self.on_error), *a.proc)

    def stdio(self, pipes: AsyncioStdio) -> IO[Coroutine]:
        proto = BasicProto(self.asio, self.on_message, self.on_error)
        async def connect() -> None:
            await self.asio.loop.connect_read_pipe(lambda: proto, sys.stdin)
            await self.asio.loop.connect_write_pipe(lambda: proto, sys.stdout)
        return IO.pure(connect())

    def socket(self, pipes: AsyncioSocket) -> IO[Coroutine]:
        return self.asio.loop.create_unix_connection(
            lambda: BasicProto(self.asio, self.on_message, self.on_error),
            str(pipes.path),
        )


def asyncio_main_loop(asio: Asyncio) -> None:
    try:
        asio.loop.run_forever()
    except Exception as e:
        log.error(e)


@do(IO[None])
def stop_asyncio_loop(asio: Asyncio) -> Do:
    yield IO.delay(asio.loop.stop)
    yield asio.thread.thread.cata(lambda t: IO.delay(t.join, 3), IO.pure(None))
    yield IO.delay(asio.loop.close)
    yield IO.delay(asio.thread.reset)


def start_processing(asio: Asyncio) -> Callable[[OnMessage, OnError], IO[None]]:
    @do(IO[None])
    def start(on_message: OnMessage, on_error: OnError) -> Do:
        thread = yield IO.fork(asyncio_main_loop, asio)
        yield IO.delay(asio.thread.update, thread)
        connect = yield connect_asyncio(asio, on_message, on_error)(asio.pipes)
        conn_finished = yield IO.delay(run_coroutine_threadsafe, connect, asio.loop)
        yield IO.delay(conn_finished.result)
    return start


def stop_processing(asio: Asyncio) -> Callable[[], IO[None]]:
    def stop() -> IO[None]:
        return stop_asyncio_loop(asio)
    return stop


# FIXME stop loop on error?
def asyncio_send(resources: AsyncioResources) -> Callable[[bytes], None]:
    def send(data: bytes) -> None:
        transport = resources.transport.result()
        Try(transport.write, data).lmap(lambda err: log.error(f'asyncio write failed: {err}'))
    return send


def join_asyncio_loop(asio: Asyncio) -> IO[None]:
    return asio.thread.thread.cata(lambda t: IO.delay(t.join), IO.failed(f'no asyncio loop running'))


def asyncio_exit(asio: Asyncio) -> None:
    pass


__all__ = ('asyncio_exit', 'join_asyncio_loop', 'asyncio_send', 'stop_processing', 'start_processing',)



from typing import TypeVar, Callable, Generic, Tuple

from amino import do, Do, List, Just
from amino.case import Case
from amino.logging import module_log
from amino.lenses.lens import lens

from ribosome.util.menu.data import (Menu, MenuAction, MenuQuit, MenuPrompt, MenuUpdateLines, MenuUnit, MenuState,
                                     MenuUpdateCursor, visible_lines, MenuStackAction, MenuStackQuit, MenuStackPush,
                                     MenuStackPop, Menus, MenuPush, MenuPop, MenuQuitWith, MenuConfig, MenuContent)
from ribosome.nvim.io.compute import NvimIO
from ribosome.util.menu.prompt.run import prompt
from ribosome.nvim.scratch import (create_scratch_buffer, CreateScratchBufferOptions, ScratchBuffer,
                                   set_scratch_buffer_content, focus_scratch_window)
from ribosome.util.menu.prompt.data import (InputChar, InputState, PromptAction, PromptUnit, PromptStateTrans,
                                            PromptConsumerUpdate, PromptQuit)
from ribosome.nvim.io.state import NS
from ribosome.nvim.api.ui import set_cursor, close_buffer, window_set_height, wincmd
from ribosome.compute.prog import Prog
from ribosome.compute.ribosome_api import Ribo
from ribosome.nvim.io.api import N
from ribosome.nvim.api.command import nvim_command

log = module_log()
S = TypeVar('S')
ML = TypeVar('ML')
U = TypeVar('U')


@do(NvimIO[None])
def cleanup_menu(scratch: ScratchBuffer) -> Do:
    yield N.recover_failure(close_buffer(scratch.buffer), lambda a: N.unit)


@do(NS[InputState[MenuState[S, ML, U], U], PromptAction])
def menu_reentry(
        handle_input: Callable[[PromptConsumerUpdate[U]], NS[InputState[MenuState[S, ML, U], U], MenuAction]],
        scratch: ScratchBuffer,
        config: MenuConfig[S, ML, U],
) -> Do:
    state = yield NS.get()
    return NS.lift(prompt(update_menu(handle_input, scratch, config), state, config.insert))


def display_lines(config: MenuConfig[S, ML, U], content: MenuContent[ML]) -> List[str]:
    lines = visible_lines(content).map(lambda a: a.text)
    return (
        lines.reversed
        if config.bottom else
        lines
    )


@do(NS[InputState[MenuState[S, ML, U], U], None])
def menu_update_cursor(bottom: bool, scratch: ScratchBuffer) -> Do:
    cursor = yield NS.inspect(lambda a: a.data.cursor)
    total = yield NS.inspect(lambda a: len(a.data.content.lines))
    line = total - cursor - 1 if bottom else cursor
    yield NS.lift(set_cursor(scratch.ui.window, (line, 0)))
    yield NS.lift(nvim_command('redraw'))


class execute_menu_action(Case[MenuAction, NS[InputState[MenuState[S, ML, U], U], PromptAction]], alg=MenuAction):

    def __init__(
            self,
            scratch: ScratchBuffer,
            handle_input: Callable[[PromptConsumerUpdate[U]], NS[InputState[MenuState[S, ML, U], U], MenuAction]],
            config: MenuConfig[S, ML, U],
    ) -> None:
        self.scratch = scratch
        self.handle_input = handle_input
        self.config = config

    def quit(self, action: MenuQuit) -> NS[InputState[MenuState[S, ML, U], U], PromptAction]:
        return NS.pure(PromptQuit())

    @do(NS[InputState[MenuState[S, ML, U], U], PromptAction])
    def quit_with(self, action: MenuQuitWith) -> Do:
        yield NS.modify(lens.data.result.set(Just(action.next)))
        return PromptQuit()

    def prompt_state_trans(self, action: MenuPrompt) -> NS[InputState[MenuState[S, ML, U], U], PromptAction]:
        return NS.pure(PromptStateTrans(action.state))

    @do(NS[InputState[MenuState[S, ML, U], U], PromptAction])
    def update_lines(self, action: MenuUpdateLines) -> Do:
        lines = display_lines(self.config, action.content)
        height = max(1, min(len(lines), self.config.max_size.get_or_strict(1000)))
        yield NS.modify(lens.data.content.set(action.content))
        yield NS.lift(window_set_height(self.scratch.ui.window, height))
        yield NS.lift(set_scratch_buffer_content(self.scratch, lines))
        yield menu_update_cursor(self.config.bottom, self.scratch)
        return PromptUnit()

    @do(NS[InputState[MenuState[S, ML, U], U], PromptAction])
    def update_cursor(self, a: MenuUpdateCursor) -> Do:
        yield menu_update_cursor(self.config.bottom, self.scratch)
        return PromptUnit()

    @do(NS[InputState[MenuState[S, ML, U], U], PromptAction])
    def push(self, action: MenuPush) -> Do:
        current = yield menu_reentry(self.handle_input, self.scratch, self.config)
        yield NS.modify(lens.data.next.set(MenuStackPush(current, action.thunk(self.scratch))))
        return PromptQuit()

    @do(NS[InputState[MenuState[S, ML, U], U], PromptAction])
    def pop(self, action: MenuPop) -> Do:
        yield NS.modify(lens.data.next.set(MenuStackPop()))
        return PromptQuit()

    def unit(self, a: MenuUnit) -> NS[InputState[MenuState[S, ML, U], U], PromptAction]:
        return NS.pure(PromptUnit())


def update_menu(
        handle_input: Callable[[PromptConsumerUpdate[U]], NS[InputState[MenuState[S, ML, U], U], MenuAction]],
        scratch: ScratchBuffer,
        config: MenuConfig[S, ML, U],
) -> Callable[[InputChar], NS[InputState[MenuState[S, ML, U], U], PromptAction]]:
    @do(NS[InputState[MenuState[S, ML, U], U], PromptAction])
    def update_menu(update: PromptConsumerUpdate[U]) -> Do:
        action = yield handle_input(update)
        yield execute_menu_action(scratch, handle_input, config)(action)
    return update_menu


class menu_recurse(Generic[S, ML], Case[MenuStackAction, NS[Menus, Prog[None]]], alg=MenuStackAction):

    def __init__(self, result: MenuState[S, ML, U]) -> None:
        self.result = result

    def quit(self, action: MenuStackQuit) -> NS[Menus, Prog[None]]:
        return NS.pure(self.result.result)

    @do(NS[Menus, Prog[None]])
    def push(self, action: MenuStackPush) -> Do:
        yield NS.modify(lambda a: a.mod.stack(lambda b: b.cons(action.current)))
        yield action.thunk

    @do(NS[Menus, Prog[None]])
    def pop(self, action: MenuStackPop) -> Do:
        stack = yield NS.inspect(lambda a: a.stack)
        def push(
                menus: Menus,
                h: NS[Menus, Prog[None]],
                t: List[Menus, Prog[None]],
        ) -> NvimIO[Tuple[Menus, NS[Menus, Prog[None]]]]:
            return N.pure((menus.set.stack(t), h))
        next = yield stack.detach_head.map2(
            lambda h, t: NS.apply(lambda a: push(a, h, t))
        ).get_or(NS.pure, NS.pure(self.result.result))
        yield next


def menu_prompt(menu: Menu[S, ML, U], scratch: ScratchBuffer) -> NvimIO[MenuState[S, ML, U]]:
    state = MenuState.cons(menu.initial_state, menu.config, MenuContent.cons(menu.lines))
    return prompt(update_menu(menu.handle_input, scratch, menu.config), state, menu.config.insert)


@do(NS[Menus, Prog[None]])
def menu_loop(menu: Menu[S, ML, U], scratch: ScratchBuffer) -> Do:
    result = yield NS.lift(menu_prompt(menu, scratch))
    yield menu_recurse(result)(result.next)


@do(NvimIO[Prog[None]])
def run_menu(menu: Menu[S, ML, U]) -> Do:
    scratch = yield create_scratch_buffer(CreateScratchBufferOptions.cons(wrap=False, name=menu.config.name))
    yield wincmd(scratch.ui.window, 'J')
    yield focus_scratch_window(scratch)
    state, next = yield N.ensure(menu_loop(menu, scratch).run(Menus.cons()), lambda a: cleanup_menu(scratch))
    return next.get_or_strict(Prog.unit).replace(state)


@do(Prog[S])
def run_menu_prog(menu: Menu[S, ML, U]) -> Do:
    next = yield Ribo.lift_nvimio(run_menu(menu))
    yield next


def menu_push(menu: Menu[S, ML, U]) -> MenuPush:
    return MenuPush(lambda scratch: menu_loop(menu, scratch))


__all__ = ('run_menu', 'run_menu_prog',)



from asyncio import new_event_loop
from typing import Tuple
from concurrent.futures import Future

from amino import List, IO, do, Do, Path
from amino.logging import module_log

from ribosome.rpc.comm import RpcComm
from ribosome.config.config import Config
from ribosome.rpc.start import start_plugin_sync, cannot_execute_request, init_comm
from ribosome.rpc.nvim_api import RiboNvimApi
from ribosome.rpc.io.data import AsyncioPipes, Asyncio, AsyncioResources, AsyncioEmbed, AsyncioStdio, AsyncioSocket
from ribosome.rpc.io.connect import start_processing, stop_processing, asyncio_send, join_asyncio_loop, asyncio_exit

log = module_log()
embed_nvim_cmdline = List('nvim', '-n', '-u', 'NONE', '--embed')


def cons_asyncio(pipes: AsyncioPipes) -> Tuple[Asyncio, RpcComm]:
    loop = new_event_loop()
    resources = AsyncioResources.cons(Future())
    asio = Asyncio.cons(loop, pipes, resources)
    comm = RpcComm(
        start_processing(asio),
        stop_processing(asio),
        asyncio_send(resources),
        lambda: join_asyncio_loop(asio),
        lambda: asyncio_exit(asio),
    )
    return asio, comm


def cons_asyncio_embed(proc: List[str]) -> Tuple[Asyncio, RpcComm]:
    return cons_asyncio(AsyncioEmbed(proc))


def cons_asyncio_stdio() -> Tuple[Asyncio, RpcComm]:
    return cons_asyncio(AsyncioStdio())


def cons_asyncio_socket(path: Path) -> Tuple[Asyncio, RpcComm]:
    return cons_asyncio(AsyncioSocket(path))


def start_asyncio_plugin_sync(config: Config) -> IO[None]:
    asio, rpc_comm = cons_asyncio_stdio()
    return start_plugin_sync(config, rpc_comm)


@do(IO[RiboNvimApi])
def start_asyncio_embed_nvim_sync(name: str, extra: List[str]) -> Do:
    asio, rpc_comm = cons_asyncio_embed(embed_nvim_cmdline + extra)
    comm = yield init_comm(rpc_comm, cannot_execute_request)
    return RiboNvimApi(name, comm)


def start_asyncio_embed_nvim_sync_log(name: str, log: Path) -> IO[RiboNvimApi]:
    return start_asyncio_embed_nvim_sync(name, List(f'-V{log}'))


__all__ = ('cons_asyncio_embed', 'cons_asyncio_stdio', 'cons_asyncio_socket', 'start_asyncio_plugin_sync',
           'start_asyncio_embed_nvim_sync', 'start_asyncio_embed_nvim_sync_log',)
